# # import asyncio
# # import logging
# # import os
# # from urllib.parse import urljoin, urlparse, urlunparse
# # from typing import List

# # import aiohttp
# # from loguru import logger


# # from tqdm.asyncio import tqdm_asyncio


# async def process_batch(batch, client: Lean4Client, timeout, infotree_type, semaphore):
#     async with semaphore:
#         response = await client.async_verify(
#             batch, timeout=timeout, infotree_type=infotree_type
#         )
#         return response


# async def process_batches(
#     client,
#     batches: List[List[dict]],
#     timeout=60,
#     num_proc=os.cpu_count(),
#     infotree_type=None,
# ):
#     semaphore = asyncio.Semaphore(num_proc)

#     results = []

#     coros = [
#         process_batch(batche, client, timeout, infotree_type, semaphore)
#         for batche in batches
#     ]

#     for fut in tqdm_asyncio.as_completed(
#         coros, total=len(batches), desc="Verifying proofs"
#     ):
#         result = await fut
#         results.extend(result["results"])

#     return results


# def batch_verify_proof(
#     client,
#     samples: List[dict],
#     timeout=60,
#     num_proc=os.cpu_count(),
#     batch_size=8,
#     infotree_type=None,
# ):
#     custom_ids = [sample["custom_id"] for sample in samples]
#     assert len(custom_ids) == len(set(custom_ids)), "Custom id must be unique"

#     logger.info(
#         f"Processing {len(samples)} samples in {len(samples)/batch_size} batches of size {batch_size}"
#     )

#     batches = [samples[i : i + batch_size] for i in range(0, len(samples), batch_size)]

#     results = asyncio.run(
#         process_batches(
#             client,
#             batches,
#             timeout=timeout,
#             num_proc=num_proc,
#             infotree_type=infotree_type,
#         )
#     )

#     return results



import os
from typing import Any, Type, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class BaseKimina:
    def __init__(
        self,
        api_url: str | None = None,
        api_key: str | None = None,
        headers: dict[str, str] | None = None,
        http_timeout: int = 60,
        n_retries: int = 3,
    ):
        if not api_url:
            api_url = os.getenv("LEAN_SERVER_API_URL", "http://localhost:8000")
        self.api_url = api_url.rstrip("/")

        if not api_key:
            api_key = os.getenv("LEAN_SERVER_API_KEY") or os.getenv(
                "LEANSERVER_API_KEY"
            )

        self.api_key = api_key
        self.headers = headers or {}
        if self.api_key:
            self.headers.setdefault("Authorization", f"Bearer {self.api_key}")
        self.http_timeout = http_timeout
        self.n_retries = n_retries

    def build_url(self, path: str) -> str:
        return f"{self.api_url}/{path.lstrip('/')}"

    def handle(self, resp: Any, model: Type[T]) -> T:
        return model.model_validate(resp)



import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

import httpx
from datasets import load_dataset, load_dataset_builder  # type: ignore
from tenacity import (
    RetryError,
    before_sleep_log,
    retry,
    stop_after_attempt,
    wait_exponential,
)
from tqdm import tqdm

from .base import BaseKimina
from .models import CheckRequest, CheckResponse, Infotree, ReplResponse, Snippet
from .utils import build_log, find_code_column, find_id_column

logger = logging.getLogger("kimina-client")


class KiminaClient(BaseKimina):
    def __init__(
        self,
        api_url: str | None = None,
        api_key: str | None = None,
        headers: dict[str, str] | None = None,
        http_timeout: int = 600,
        n_retries: int = 3,
    ):
        super().__init__(
            api_url=api_url,
            api_key=api_key,
            headers=headers,
            http_timeout=http_timeout,
            n_retries=n_retries,
        )

    def check(
        self,
        snips: str | list[str] | Snippet | list[Snippet],
        timeout: int = 60,
        debug: bool = False,
        reuse: bool = True,
        infotree: Infotree | None = None,
        batch_size: int = 8,
        max_workers: int = 5,
        show_progress: bool = True,
    ) -> CheckResponse:
        if isinstance(snips, str):
            snips = [snips]
        elif isinstance(snips, Snippet):
            snips = [snips]

        snippets = [Snippet.from_snip(snip) for snip in snips]
        batches = [
            snippets[i : i + batch_size] for i in range(0, len(snippets), batch_size)
        ]
        results: list[CheckResponse] = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {
                executor.submit(
                    self.api_check, batch, timeout, debug, reuse, infotree, True
                ): batch
                for batch in batches
            }
            iterator = (
                tqdm(
                    as_completed(futures),
                    total=len(futures),
                    desc="Batches",
                    bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [elapsed: {elapsed}, remaining: {remaining}]",
                )
                if show_progress
                else as_completed(futures)
            )
            for future in iterator:
                results.append(future.result())
        return CheckResponse.merge(results)

    def api_check(
        self,
        snippets: list[Snippet],
        timeout: int = 30,
        debug: bool = False,
        reuse: bool = True,
        infotree: Infotree | None = None,
        safe: bool = False,
    ) -> CheckResponse:
        """
        Makes a POST request to /api/check with provided arguments.

        Returns a `CheckResponse`.
        """
        try:
            url = self.build_url("/api/check")

            payload = CheckRequest(
                snippets=snippets,
                timeout=timeout,
                debug=debug,
                reuse=reuse,
                infotree=infotree,
            ).model_dump()

            resp = self._query(url, payload)
            return self.handle(resp, CheckResponse)
        except Exception as e:
            if safe:
                return CheckResponse(
                    results=[
                        ReplResponse(id=snip.id, error=str(e)) for snip in snippets
                    ],
                )
            raise e

    def _query(
        self, url: str, payload: dict[str, Any] | None = None, method: str = "POST"
    ) -> Any:
        """
        Sends a `method` request to `url` with `payload` as body/params.
        A new `httpx.Client` is created for each request for thread-safety.
        Use AsyncClient for more efficient concurrent requests (TCP connection reuse/pooling).
        """

        @retry(
            stop=stop_after_attempt(self.n_retries),
            wait=wait_exponential(multiplier=1, min=1, max=10),
            before_sleep=before_sleep_log(logger, logging.ERROR),
        )
        def run_method() -> Any:
            try:
                with httpx.Client(
                    headers=self.headers, timeout=self.http_timeout
                ) as client:
                    if method.upper() == "POST":
                        response = client.post(url, json=payload)
                    elif method.upper() == "GET":
                        response = client.get(url, params=payload)
                    else:
                        raise ValueError(f"Unsupported method: {method}")
                    response.raise_for_status()  # Ensure 2xx, otherwise retry
            except httpx.HTTPError as e:
                logger.error(f"Error posting to {url}: {e}")
                raise e

            try:
                return response.json()  # Ensure JSON, otherwise retry
            except ValueError:
                logger.error(f"Server returned non-JSON: {response.text}")
                raise ValueError("Invalid response from server: not a valid JSON")

        try:
            return run_method()
        except RetryError:
            raise RuntimeError(f"Request failed after {self.n_retries} retries")

    def health(self) -> Any:
        """
        Checks server's healthy.
        """
        url = self.build_url("/health")
        resp = self._query(url, method="GET")
        return resp  # TODO: create status object to cast automatically

    def test(self) -> None:
        """
        Tests with `#check Nat`.
        """
        logger.info("Testing with `#check Nat`...")
        response = self.check("#check Nat", show_progress=False).results[0].response
        assert response is not None, "Response should not be None"
        assert response.get("messages", None) == [
            {
                "severity": "info",
                "pos": {"line": 1, "column": 0},
                "endPos": {"line": 1, "column": 6},
                "data": "Nat : Type",
            }
        ]
        logger.info("Test passed!")

    def run_benchmark(
        self,
        dataset_name: str = "Goedel-LM/Lean-workbook-proofs",
        split: str = "train",
        n: int = 100,
        batch_size: int = 8,
        max_workers: int = 5,
        timeout: int = 60,
        reuse: bool = True,
        show_progress: bool = True,
    ) -> None:
        """
        Runs benchmark on Hugging Face dataset.
        Displays results in the console.
        """
        # TODO: add option output dir with file hierarchy based on metadata like uuid in the run_benchmark method
        # TODO: add count heartbeats option

        if n <= 0:
            logger.error("Please specify n > 0")
            return

        if batch_size <= 0:
            logger.warning("Cannot use batch size = %d, defaulting to 8", batch_size)
            batch_size = 8

        logger.info(build_log(dataset_name, n, batch_size))

        builder = load_dataset_builder(dataset_name)
        if not builder.info.features:
            logger.error("Dataset has no features, cannot run benchmark")
            return

        columns: list[str] = list(builder.info.features)

        id_column_name: str | tuple[str, str] = "id"
        code_column_name: str = "code"
        if dataset_name == "Goedel-LM/Lean-workbook-proofs":
            id_column_name = "problem_id"
            code_column_name = "full_proof"
        elif dataset_name == "AI-MO/math-test-inference-results":
            id_column_name = ("uuid", "proof_id")
            code_column_name = "proof"
        else:
            id_column_name = find_id_column(columns)
            code_column_name = find_code_column(columns)

        dataset = load_dataset(dataset_name, split=split + f"[:{n}]")

        def get_id(sample: Any, id_column_name: str | tuple[str, str]) -> str:
            if isinstance(id_column_name, tuple):
                a, b = id_column_name
                return str(sample[a]) + "_" + str(sample[b])
            return str(sample[id_column_name])

        snips = [
            Snippet(
                id=str(get_id(sample, id_column_name)),
                code=sample[code_column_name],  # type: ignore
            )
            for sample in dataset  # type: ignore
        ]

        start_time = time.time()
        check_response = self.check(
            snips=snips,
            timeout=timeout,
            reuse=reuse,
            batch_size=batch_size,
            max_workers=max_workers,
            show_progress=show_progress,
        )
        elapsed_time = time.time() - start_time

        check_response.analyze(elapsed_time)



import json
import multiprocessing as mp
import time
import pandas as pd
from datasets import load_dataset
from loguru import logger
from tqdm import tqdm

from autoformalizer.eval_utils import lean_feedback


async def verify_one_batch(client, timeout, working_dir, infotree_type, samples):
    """
        - samples (list): A list of samples to verify in one batch
            samples must contain the following keys:
                - proof: The proof code to verify
                - proof_id: The proof id
                - uuid: The uuid of the problem
        - timeout (int): The timeout in seconds for the entire batch
        - working_dir (Path): Working directory for saving verification results

    Returns:
        - list: List of processed samples with verification results
    """
    # Prepare batch input format
    batch_codes = []
    for sample in samples:
        batch_codes.append(
            {"code": sample["proof"], "custom_id": str(sample["proof_id"])}
        )
    id_to_sample = {str(sample["proof_id"]): sample for sample in samples}

    # Perform batch verification
    start_time = time.time()
    if working_dir:
        # Create directory structure for results
        for sample in samples:
            uuid_dir = working_dir / sample["uuid"]
            uuid_dir.mkdir(parents=True, exist_ok=True)

        # check if all file path exist
        all_file_path = [
            working_dir / sample["uuid"] / f"{sample['proof_id']}.json"
            for sample in samples
        ]
    if working_dir and all([file_path.exists() for file_path in all_file_path]):
        # all files exist, load reponse
        response = []
        for file_path in all_file_path:
            with open(file_path, "r") as f:
                try:
                    data = json.load(f)
                except json.JSONDecodeError:
                    data = {"error": "JSONDecodeError"}
                if "code" not in data:
                    data["code"] = file_path.stem
                response.append(data)
    else:
        response = await client.async_one_pass_verify_batch(
            batch_codes, timeout=timeout, infotree_type=infotree_type
        )
        response = response.get("results", [])
    elapsed_time = time.time() - start_time

    # Process results
    if len(response) != len(samples):
        logger.error(
            f"""Time: {elapsed_time:.2f}s, Batch verification failed:
            Expected {len(samples)} results, got {len(response)}"""
        )
        # return samples with default values
        results = []
        for sample in samples:
            new_sample = sample.copy()
            new_sample.update(
                {
                    "lean_feedback": json.dumps(
                        {"error": "Missing response from server"}
                    ),
                    "has_error": True,
                    "is_valid_no_sorry": False,
                    "is_valid_with_sorry": False,
                }
            )
            results.append(new_sample)
        return results

    results = []
    for individual_response in response:
        if individual_response.get("error") is not None:
            # ignore timeout error
            if "Lean process timed out" not in individual_response["error"]:
                logger.warning(
                    f"Time: {elapsed_time:.2f}s, Sample in batch failed: {individual_response['error']}"
                )
        elif individual_response.get("response") is None:
            logger.warning(
                f"Time: {elapsed_time:.2f}s, Sample in batch failed: No response found"
            )

        proof_id = (
            individual_response["custom_id"]
            if "custom_id" in individual_response
            else individual_response["code"]
        )
        sample = id_to_sample.get(proof_id).copy()
        filepath = (
            working_dir / sample["uuid"] / f"{proof_id}.json" if working_dir else None
        )
        if working_dir and (not filepath.exists()):
            with open(filepath, "w") as f:
                json.dump(individual_response, f)

        output = lean_feedback.parse_client_response(individual_response)
        sample.update(
            {
                "lean_feedback": json.dumps(individual_response),
                "has_error": output["has_error"],
                "is_valid_no_sorry": output["is_valid_no_sorry"],
                "is_valid_with_sorry": output["is_valid_with_sorry"],
            }
        )
        results.append(sample)

    return results


# TODO: no count heartbeats yet
async def thread_worker(
    batch_queue, result_queue, client, timeout, working_dir, infotree_type
):
    results = await verify_one_batch(
        client, timeout, working_dir, infotree_type, batch
    )
    except Exception as e:
        logger.error(
            f"Batch verification failed for batch {str(batch)[:50]} with error: {e}"
        )
        results = []
        for sample in batch:
            new_sample = sample.copy()
            new_sample.update(
                {
                    "lean_feedback": json.dumps({"error": str(e)}),
                    "has_error": True,
                    "is_valid_no_sorry": False,
                    "is_valid_with_sorry": False,
                }
            )
            results.append(new_sample)
    result_queue.put(results)

def batch_verify_proof(
    client,
    samples,
    timeout=60,
    num_threads=20,
    batch_size=8,
    working_dir=None,
    infotree_type=None,
    disable_tqdm=False,
    count_heartbeats=False,
):
    """
        - samples (list): List of samples to verify.
            Must contain the following keys:
                - proof_id: The proof id
                - uuid: The uuid of the problem
                - proof: The formal proof to verify
        - working_dir (Path): Working directory for saving verification results
            if None, no cache is applied

    Returns:
        - list: List of all processed samples with verification results
            Each sample contains the following:
                - proof_id: The proof id
                - uuid: The uuid of the problem
                - proof: The formal proof
                - lean_feedback: A JSON string with the response from the server
                - has_error: A boolean indicating if the proof has an error
                - is_valid_no_sorry: A boolean indicating if the proof is valid
                    without any "sorry" statements.
                    This is the output from lean_feedback.parse_client_response.
                - is_valid_with_sorry: A boolean indicating if the proof is valid
                    with "sorry" statements. Use to check formal_statement.
                    This is the output from lean_feedback.parse_client_response.
                - heartbeats: count of heartbeats for all theorems

    """

    # assert proof id is unique
    proof_ids = [sample["proof_id"] for sample in samples_copy]
    assert len(proof_ids) == len(set(proof_ids)), "Proof id must be unique"

    manager = mp.Manager()
    batch_queue = manager.Queue()
    result_queue = manager.Queue()
    [batch_queue.put(batch) for batch in batches]

    all_results = []
    
    # count heartbeats
    for data in all_results:
        tot_heartbeats = extract_all_heartbeats_lean_output(data)
        data["heartbeats"] = tot_heartbeats

    return all_results



def example_batch(client):
    dataset = load_dataset("AI-MO/math-test-inference-results", split="train")
    working_dir = None
    # select the first 500 uuids
    dataset = dataset.select(range(100 * 64))
    print(dataset)

    # make sure we have proof_id, uuid, and proof
    new_samples = []
    for sample in tqdm(dataset):
        new_samples.append(
            {
                # for this dataset, proof_id is not unique, unfortunately
                "proof_id": sample["uuid"] + "_" + str(sample["proof_id"]),
                "uuid": sample["uuid"],
                "proof": sample["proof"],
            }
        )

    # create batch and use one_pass_verify_batch
    batch_size = 1
    results = batch_verify_proof(
        client=client,
        samples=new_samples,
        timeout=60,
        num_threads=200,
        batch_size=batch_size,
        working_dir=working_dir,
        count_heartbeats=True,
    )
    # print(len(results))
    # print(results[0])

    # calculate valid rate and other stats
    df = []
    for data in results:
        uuid = data["uuid"]
        name = data["proof_id"]
        response = json.loads(data["lean_feedback"])
        error_message = response.get("error", None)
        df.append(
            {
                "uuid": uuid,
                "is_valid_no_sorry": data["is_valid_no_sorry"],
                "name": name,
                "has_connection_error": bool(error_message),
                "heartbeats": data["heartbeats"],
            }
        )
    df = pd.DataFrame(df)
    print(df)

    # calculate valid rate
    valid_rate = df["is_valid_no_sorry"].sum() / len(df)
    print(f"valid rate: {valid_rate}")

    # connection error rate
    connection_error_rate = df["has_connection_error"].sum() / len(df)
    print(f"connection error rate: {connection_error_rate}")

    # calculate valid rate for each uuid using groupby
    uuid_group = df.groupby("uuid")["is_valid_no_sorry"].sum()

    # find all uuids with at least one valid proof
    valid_uuids = uuid_group[uuid_group > 0].index
    print(f"Number of uuids: {len(uuid_group)}")
    print(f"Number of uuids with at least one valid proof: {len(valid_uuids)}")
    if len(valid_uuids) == 52:
        logger.info("Example batch test passed!")
    else:
        logger.error("Example batch test failed!")
        raise ValueError("Example batch test failed!")


# Not used anywhere, let's convert it to a sanity check against the server (test connection++)
def test_single_proof(client):
    """
    Tests a single proof with the Lean4 client.

    Args:
        client (Lean4Client): The client to use for verification.
        proof (str): The proof code to verify.

    Returns:
        dict: The response from the server.
    """
    proof = """import Mathlib
import Aesop

open BigOperators Real Nat Topology Rat

set_option maxHeartbeats 0

/-- The volume of a cone is given by the formula $V = \frac{1}{3}Bh$,
where $B$ is the area of the base and $h$ is the height. The area of the base of a cone is 30 square units,
and its height is 6.5 units. What is the number of cubic units in its volume? Show that it is 65.-/
theorem mathd_algebra_478 (b h v : ℝ) (h₀ : 0 < b ∧ 0 < h ∧ 0 < v) (h₁ : v = 1 / 3 * (b * h))
    (h₂ : b = 30) (h₃ : h = 13 / 2) : v = 65 := by
  -- Start with the formula for the volume of the cone
  rw [h₁, h₂, h₃] -- Substitute the values for b and h
  -- Simplify the expression: V = (1/3) * 30 * (13/2)
  norm_num
"""

    code = {
        "uuid": "aaa",
        "proof_id": "aaa_0",
        "proof": proof,
    }
    samples = [code]
    res = batch_verify_proof(
        client, samples=samples, timeout=15, num_threads=3, count_heartbeats=True
    )
    logger.info(f"code sample after verify :\n{samples[0]['proof']}")

    return res


if __name__ == "__main__":
    """
    python -m autoformalizer.clients.lean4_client
    Example batch should display something like

    ==========================Output==========================
        valid rate: 0.32890625
        connection error rate: 0.0
        Number of uuids: 100
        Number of uuids with at least one valid proof: 52
    """
    client = Lean4Client()

    if test_connection(client):
        logger.info("Connection successful!")
    else:
        logger.error("Connection failed!")

    output = test_single_proof(client)
    print(output[0]["lean_feedback"])
    if output[0]["is_valid_no_sorry"]:
        logger.info("Proof is valid!")
        heartbeats = extract_first_heartbeats_lean_output(output[0])
        logger.info(f"heartbeats : {heartbeats}")
    else:
        logger.error("Proof is invalid!")

    # example_batch(client)


# import re
# from typing import Optional


# def extract_nodes_and_edges(
#     infotree: list[dict],
#     parent_id: Optional[int] = None,
#     start_id: int = 0,
#     include_failed_pp: bool = True,
#     deduplicate: bool = False,
# ):
#     """
#     Recursively extract nodes and edges from an infotree.

#     Parameters
#     ----------
#     infotree : list of dict
#         A list of dictionaries each containing 'node' and optionally 'children'.
#     parent_id : int or None, optional
#         The ID of the parent node, or None if this is the root level (default is None).
#     start_id : int, optional
#         The next available integer ID to assign to a new node (default is 0).
#     include_failed_pp : bool, optional
#         If False, nodes whose pretty-print text is "<failed to pretty print>" are removed
#         (default is True).
#     deduplicate : bool, optional
#         If True, deduplicate chains of identical nodes based on goalsBefore,
#         goalsAfter, and pp (default is False).

#     Returns
#     -------
#     nodes : dict[int, dict]
#         A dictionary of node_id -> node_content.
#     edges : list[tuple[int, int, dict]]
#         A list of tuples (parent_id, child_id, {}) representing edges in the infotree.
#     next_id : int
#         The next available integer ID after processing all children.
#     """
#     nodes = {}
#     edges = []

#     current_id = start_id

#     for item in infotree:
#         if "node" in item:
#             node_data = item["node"]

#             # Add this node to the nodes dictionary
#             node_id = current_id
#             current_id += 1
#             nodes[node_id] = node_data

#             # Recursively handle the children of the node, depth-first approach
#             if (
#                 "children" in item
#                 and isinstance(item["children"], list)
#                 and item["children"]
#             ):
#                 child_nodes, child_edges, current_id = extract_nodes_and_edges(
#                     item["children"],
#                     parent_id=node_id,
#                     start_id=current_id,
#                     include_failed_pp=include_failed_pp,
#                     deduplicate=deduplicate,
#                 )
#                 nodes.update(child_nodes)
#                 edges.extend(child_edges)

#             # Add an edge from the parent to this node
#             if parent_id is not None:
#                 edges.append((parent_id, node_id, {}))

#             # Now handle possible flattening in a loop until there are no more changes
#             transformed = True
#             while transformed:
#                 transformed = False

#                 # 1) Remove/flatten all children that have failed PP (if include_failed_pp=False)
#                 #    We handle them individually, even if there's more than one child.
#                 #    Then we break to re-check children from scratch, as new failed PP children might appear.
#                 child_list = [
#                     e for e in edges if e[0] == node_id
#                 ]  # edges from node_id to child
#                 for edge_obj in child_list:
#                     child_id = edge_obj[1]
#                     child_content = nodes.get(child_id, {})
#                     child_pp = child_content.get("stx", {}).get("pp", "")

#                     if not include_failed_pp and child_pp == "<failed to pretty print>":
#                         # Flatten this child: remove it and connect parent directly to its children
#                         # Note that the new children might also have failed PP, so we need to re-check them
#                         _flatten_chain(nodes, edges, node_id, child_id)
#                         transformed = True
#                         break

#                 if transformed:
#                     # We need to restart the while loop to re-check children from scratch, as there might be
#                     # more failed PP children
#                     continue

#                 # 2) Deduplicate if there's exactly one child left that has same (goalsBefore, goalsAfter, pp)
#                 # This happens quite often in infotrees, where a tactic is repeated multiple times, extracted
#                 # with different parsers
#                 child_list = [e for e in edges if e[0] == node_id]
#                 if deduplicate and len(child_list) == 1:
#                     child_id = child_list[0][1]
#                     if child_id in nodes:
#                         child_content = nodes[child_id]
#                         # Compare parent's vs child's (goalsBefore, goalsAfter, pp)
#                         parent_goalsBefore = node_data.get("goalsBefore", [])
#                         parent_goalsAfter = node_data.get("goalsAfter", [])
#                         parent_pp = node_data.get("stx", {}).get("pp", "")

#                         child_goalsBefore = child_content.get("goalsBefore", [])
#                         child_goalsAfter = child_content.get("goalsAfter", [])
#                         child_pp = child_content.get("stx", {}).get("pp", "")

#                         if (
#                             child_goalsBefore == parent_goalsBefore
#                             and child_goalsAfter == parent_goalsAfter
#                             and child_pp == parent_pp
#                         ):
#                             # Flatten this child: remove it and connect parent directly to its children
#                             _flatten_chain(nodes, edges, node_id, child_id)
#                             transformed = True

#         else:
#             # If the item does not contain a 'node' key but might have children
#             if "children" in item and isinstance(item["children"], list):
#                 child_nodes, child_edges, current_id = extract_nodes_and_edges(
#                     item["children"],
#                     parent_id=parent_id,
#                     start_id=current_id,
#                     include_failed_pp=include_failed_pp,
#                     deduplicate=deduplicate,
#                 )
#                 nodes.update(child_nodes)
#                 edges.extend(child_edges)

#     return nodes, edges, current_id


# def _flatten_chain(nodes: dict, edges: list[tuple], parent_id: int, child_id: int):
#     """
#     Flatten a chain by removing 'child_id' node and connecting 'parent_id'
#     directly to the child's children.
#     Given a parent node that has a single child, this function removes the child
#     node from the dictionary of nodes and reassigns the child's children to the parent.
#     This is used for node deduplication and removing failed-pp nodes.

#     Parameters
#     ----------
#     nodes : dict[int, dict]
#         A dictionary of node_id -> node_content.
#     edges : list[tuple[int, int, dict]]
#         A list of tuples (parent_id, child_id, {}) representing edges in the infotree.
#     parent_id : int
#         The ID of the parent node.
#     child_id : int
#         The ID of the child node that should be removed.

#     Returns
#     -------
#     None
#         This function modifies the nodes and edges in place.
#     """
#     if child_id not in nodes:
#         return

#     # Remove the node from the dictionary
#     del nodes[child_id]

#     # Remove edge from parent_id -> child_id
#     edges[:] = [e for e in edges if not (e[0] == parent_id and e[1] == child_id)]

#     # Reassign child's children edges to the parent
#     for i, (src, tgt, attr) in enumerate(edges):
#         if src == child_id:
#             edges[i] = (parent_id, tgt, attr)


# def get_intervals(nodes: dict) -> list[dict]:
#     """
#     Build a list of intervals from a given nodes dictionary.
#     Each interval represents a tactic in the Lean file, capturing its
#     start and finish positions, as well as the associated goals.

#     Parameters
#     ----------
#     nodes : dict of {int : dict}
#         A dictionary of node_id -> node_content.

#     Returns
#     -------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           node_id, pp, start_line, start_col, finish_line, finish_col, goalsBefore, goalsAfter
#     """
#     intervals = []
#     for node_id, node_content in nodes.items():
#         stx_range = node_content.get("stx", {}).get("range", {})
#         start_dict = stx_range.get("start", {})
#         finish_dict = stx_range.get("finish", {})

#         intervals.append(
#             {
#                 "node_id": node_id,
#                 "pp": node_content.get("stx", {}).get("pp", ""),
#                 "start_line": start_dict.get("line", 0),
#                 "start_col": start_dict.get("column", 0),
#                 "finish_line": finish_dict.get("line", 0),
#                 "finish_col": finish_dict.get("column", 0),
#                 "goalsBefore": node_content.get("goalsBefore", []),
#                 "goalsAfter": node_content.get("goalsAfter", []),
#             }
#         )
#     return intervals


# def adjust_intervals(intervals: list[dict]) -> list[dict]:
#     """
#     Make intervals disjoint and create a file partition.
#     Sort intervals by starting position, then set each interval's end to the next
#     interval's start. This creates a sequence of adjacent intervals covering the file.

#     Parameters
#     ----------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           node_id, pp, start_line, start_col, finish_line, finish_col, goalsBefore, goalsAfter

#     Returns
#     -------
#     intervals_sorted : list of dict
#         The updated intervals, sorted and trimmed so that they do not overlap.
#     """
#     intervals_sorted = sorted(
#         intervals, key=lambda iv: (iv["start_line"], iv["start_col"])
#     )

#     # Remember the furthest original finish position
#     max_finish_line, max_finish_col = -1, -1
#     for iv in intervals_sorted:
#         if (iv["finish_line"], iv["finish_col"]) > (max_finish_line, max_finish_col):
#             max_finish_line, max_finish_col = iv["finish_line"], iv["finish_col"]

#     for i in range(len(intervals_sorted) - 1):
#         current = intervals_sorted[i]
#         nxt = intervals_sorted[i + 1]
#         current["finish_line"] = nxt["start_line"]
#         current["finish_col"] = nxt["start_col"]
#         current["goalsAfter"] = nxt["goalsBefore"]

#     if intervals_sorted:
#         intervals_sorted[-1]["finish_line"] = max_finish_line
#         intervals_sorted[-1]["finish_col"] = max_finish_col

#     intervals_sorted = [
#         iv
#         for iv in intervals_sorted
#         if not (
#             iv["start_line"] == iv["finish_line"]
#             and iv["start_col"] == iv["finish_col"]
#         )
#     ]

#     return intervals_sorted


# def retrieve_tactics(intervals: list[dict], source_lines: list[str]) -> list[dict]:
#     """
#     Extract tactic code snippets from source lines based on intervals.

#     Parameters
#     ----------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           node_id, pp, start_line, start_col, finish_line, finish_col, goalsBefore, goalsAfter
#         Note: At this point, the pp field does not exactly correspond to the positions.
#     source_lines : list of str
#         The lines of the Lean file, read into a list.

#     Returns
#     -------
#     results : list of dict
#         A list of intervals augmented with the 'tactic' text from the file.
#         Each dict has keys: goalsBefore, goalsAfter, tactic.
#     """
#     results = []
#     for i in range(len(intervals)):
#         iv = intervals[i]
#         snippet_text = _extract_snippet(
#             source_lines,
#             iv["start_line"],
#             iv["start_col"],
#             iv["finish_line"],
#             iv["finish_col"],
#         )
#         data = {
#             "goalsBefore": iv["goalsBefore"],
#             "goalsAfter": iv["goalsAfter"],
#             "tactic": snippet_text,
#         }
#         results.append(data)

#     return results


# def _extract_snippet(
#     source_lines: list[str],
#     start_line: int,
#     start_col: int,
#     finish_line: int,
#     finish_col: int,
# ) -> str:
#     """
#     Extract a code snippet from the Lean source lines.

#     Given a start and finish line-column pair, slice the lines to produce the exact text
#     range in the Lean file. This handles both single-line and multi-line cases.

#     Parameters
#     ----------
#     source_lines : list of str
#         The lines read from the Lean file.
#     start_line : int
#         The 1-based starting line index.
#     start_col : int
#         The 0-based starting column index within start_line.
#     finish_line : int
#         The 1-based finishing line index.
#     finish_col : int
#         The 0-based finishing column index within finish_line.

#     Returns
#     -------
#     str
#         The extracted snippet from the Lean file, spanning (start_line, start_col)
#         to (finish_line, finish_col).
#     """
#     # Single line case
#     if start_line == finish_line:
#         line_idx = start_line - 1
#         line_text = source_lines[line_idx]
#         return line_text[start_col:finish_col]

#     # Multi-line case
#     # 1) from start_col to end-of-line for start_line
#     snippet_parts = []
#     start_line_idx = start_line - 1
#     line_text = source_lines[start_line_idx]
#     snippet_parts.append(line_text[start_col:])

#     # 2) full lines between (start_line+1) .. (finish_line-1)
#     for line_idx in range(start_line_idx + 1, finish_line - 1):
#         snippet_parts.append(source_lines[line_idx])

#     # 3) from begin-of-line up to finish_col for finish_line
#     last_line_idx = finish_line - 1
#     last_line = source_lines[last_line_idx]
#     snippet_parts.append(last_line[:finish_col])

#     return "".join(snippet_parts)


# def separate_trailing_whitespace(s: str) -> tuple[str, str]:
#     """
#     Remove trailing whitespace from a tactic and return (code, trailing_ws).

#     Example:
#     - Input: "have h1 : ... := by\n    "
#     - Output: ("have h1 : ... := by", "\n    ")

#     Parameters
#     ----------
#     s : str
#         A tactic string (Lean4 code).

#     Returns
#     -------
#     code : str
#         The tactic string without potential whitespaces at the end.
#     trailing_ws : str
#         The whitespaces at the end of s. Potentially None.
#     """
#     code = s.rstrip(" \t\n\r")
#     trailing_ws = s[len(code) :]
#     return code, trailing_ws


# def separate_trailing_comment(s: str) -> tuple[str, str]:
#     """
#     Remove trailing comments only if they start at the beginning of the line
#     (modulo whitespace). Return (clean, trailing_comment).

#     A trailing comment can be:
#       - One or more single-line comments (lines starting with '--').
#       - One multi-line block comment starting with '/-' (at line start, ignoring whitespace)
#         and ending somewhere before the end of the string (must contain '-/').

#     Example:
#     - Input: "have h1 : ... := by\n  -- To prove this have statement, we will just apply mul_pos"
#     - Output: ("have h1 : ... := by\n  ", "-- To prove this have statement, we will just apply mul_pos")

#     Note that the function does not extract trailing whitespaces from the tactic right before the start of the comment.

#     Parameters
#     ----------
#     s : str
#         A tactic string (Lean4 code).

#     Returns
#     -------
#     clean : str
#         The tactic string without potential comments at the end.
#     comment : str
#         The comment at the end of s. Potentially None.
#     """
#     lines = s.splitlines(keepends=True)

#     # If no lines, do nothing
#     if not lines:
#         return s, ""

#     trailing_comments = []  # This will contain all the trailing comments
#     i = len(lines) - 1  # Start from the end of the file

#     while i >= 0:
#         line = lines[i]

#         # 1. If the line is blank, add it to the trailing_comments and go to the next line
#         if line.strip() == "":
#             trailing_comments.append(line)
#             i -= 1
#             continue

#         # 2. Check if the line ends a multi-line block comment (i.e., has '-/'):
#         if "-/" in line:
#             end_idx = i
#             start_idx = end_idx
#             # Move upward until we find a line that starts with '/-' (mod whitespace)
#             while start_idx >= 0 and not re.match(r"^\s*/-", lines[start_idx]):
#                 start_idx -= 1
#             if start_idx < 0:
#                 # Found '-/' but no valid start => not truly trailing
#                 break
#             # Everything from start_idx..end_idx is the trailing block comment
#             block_str = "".join(lines[start_idx : end_idx + 1])
#             trailing_comments.append(block_str)
#             i = start_idx - 1  # jump above the block
#             continue

#         # 3. Check if the line is a single-line comment
#         if re.match(r"^\s*--", line) is not None:
#             trailing_comments.append(line)
#             i -= 1
#             continue

#         # 4. If none of the above, then we’ve hit a real line of code
#         break

#     clean_code = "".join(lines[: i + 1])

#     # If we didn't find any trailing comment, return the code with an empty string
#     if trailing_comments == []:
#         return clean_code, ""

#     else:
#         trailing_comments.reverse()  # trailing_comments is in bottom-to-top order, reverse it to restore top-to-bottom
#         trailing_comments = "".join(trailing_comments)
#         return clean_code, trailing_comments


# def transfer_trailing_whitespaces_and_comments(intervals: list[dict]):
#     """
#     For each tactic in 'intervals' (except the last), remove trailing whitespaces,
#     then trailing comments (if any), then trailing whitespaces again,
#     and prepend all that to the next interval.

#     Parameters
#     ----------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           goalsBefore, goalsAfter, tactic.

#     Returns
#     -------
#     None
#         The 'tactic' fields in intervals are modified in place.
#     """
#     for i in range(len(intervals) - 1):
#         current = intervals[i]
#         nxt = intervals[i + 1]

#         # 1) Remove trailing whitespace
#         code, trailing_ws_1 = separate_trailing_whitespace(current["tactic"])

#         # 2) Remove trailing comment
#         code, trailing_comment = separate_trailing_comment(code)

#         # 3) Remove trailing whitespace again
#         code, trailing_ws_2 = separate_trailing_whitespace(code)

#         # Update the current tactic
#         current["tactic"] = code

#         # Prepend to the next interval
#         nxt["tactic"] = trailing_ws_2 + trailing_comment + trailing_ws_1 + nxt["tactic"]


# def remove_lean_comments(text) -> str:
#     """
#     Remove single-line and multi-line comments from `text`.

#     Parameters
#     ----------
#     text : str
#         The Lean code that may contain comments.

#     Returns
#     -------
#     text : str
#         The Lean code without comments.
#     """
#     # First, remove all multi-line comments
#     pattern = re.compile(r"/-.*?-/", re.DOTALL)
#     text = pattern.sub("", text)

#     # Then, remove all single-line comments
#     lines = text.splitlines()
#     new_lines = []
#     for line in lines:
#         if line.strip() == "":
#             new_lines.append(line)
#             continue
#         if "--" in line:
#             # Keep only the part before the first occurrence of "--"
#             line = line.split("--")[0].rstrip()
#         if line.strip() != "":
#             new_lines.append(line)
#     text = "\n".join(new_lines)

#     return text


# def is_balanced(tactic: str) -> bool:
#     """
#     Check whether `[]`, `()` and `⟨⟩` are balanced in `tactic`.

#     Parameters
#     ----------
#     tactic : str
#         A Lean tactic snippet.

#     Returns
#     -------
#     bool
#         True if every opening bracket/parenthesis is matched by its closing counterpart, False otherwise.
#     """
#     return (
#         tactic.count("[") == tactic.count("]")
#         and (tactic.count("(") == tactic.count(")"))
#         and (tactic.count("⟨") == tactic.count("⟩"))
#     )


# def is_by(tactic: str) -> bool:
#     """
#     Check whether the tactic is exactly the keyword `by`.
#     Comments are removed and surrounding whitespace is ignored before the comparison.

#     Parameters
#     ----------
#     tactic : str
#         A Lean tactic snippet.

#     Returns
#     -------
#     bool
#         True if the tactic is a 'by', False otherwise.
#     """
#     return remove_lean_comments(tactic).strip() == "by"


# def is_calc(tactic: str) -> bool:
#     """
#     Check whether the snippet introduces a `calc` block.
#     Accepted forms are either `calc` or `by calc`.

#     Parameters
#     ----------
#     tactic : str
#         A Lean tactic snippet.

#     Returns
#     -------
#     bool
#         True if the tactic introduces a `calc` block, False otherwise.
#     """
#     s = remove_lean_comments(tactic).strip()
#     # Direct match
#     if s == "calc":
#         return True
#     # Match `by calc` with spaces or newlines after `by`
#     m = re.match(r"^by\s+([\w_]+)$", s)
#     return bool(m and m.group(1) == "calc")


# WRAPPER_TACTICS: set[str] = {
#     "all_goals",
#     "any_goals",
#     "repeat",
# }


# def is_wrapper(tactic: str) -> bool:
#     """
#     Check whether the tactic is a wrapper tactic (e.g., `all_goals`, `any_goals`, `repeat`),
#     optionally preceded by `by`.

#     Parameters
#     ----------
#     tactic : str
#         A Lean tactic snippet.

#     Returns
#     -------
#     bool
#         True if the tactic is a wrapper tactic, False otherwise.
#     """
#     s = remove_lean_comments(tactic).strip()
#     # Direct match
#     if s in WRAPPER_TACTICS:
#         return True
#     # Match `by <wrapper>` with spaces or newlines after `by`
#     m = re.match(r"^by\s+([\w_]+)$", s)
#     return bool(m and m.group(1) in WRAPPER_TACTICS)


# def ends_with_by(tactic: str) -> bool:
#     """
#     Check whether the tactic ends with `:= by` or `:=by`, ignoring trailing comments and whitespace.

#     Parameters
#     ----------
#     tactic : str
#         A Lean tactic snippet.

#     Returns
#     -------
#     bool
#         True if the tactic ends with `:= by` or `:=by`, False otherwise.
#     """
#     s = remove_lean_comments(tactic).rstrip()
#     return s.endswith(":=by") or s.endswith(":= by")


# def merge_intervals(intervals: list[dict]) -> list[dict]:
#     """
#     Merge intervals that are not balanced or contain specific tactics.

#     Parameters
#     ----------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           goalsBefore, goalsAfter, tactic.

#     Returns
#     -------
#     merged_intervals : list of dict
#         A list of dictionaries, each containing:
#           goalsBefore, goalsAfter, tactic.
#         The tactics that do not change the goals are merged with their successor.
#     """
#     merged_intervals = []
#     i = 0
#     while i < len(intervals):
#         accumulated = intervals[i]["tactic"]
#         j = i + 1
#         # Merge subsequent intervals until accumulated tactic is balanced.
#         while j < len(intervals) and (
#             not is_balanced(accumulated)
#             or is_by(accumulated)
#             or is_calc(accumulated)
#             or is_wrapper(accumulated)
#         ):
#             accumulated += intervals[j]["tactic"]
#             j += 1

#         merged_interval = {
#             "goalsBefore": intervals[i]["goalsBefore"],
#             "goalsAfter": intervals[j - 1]["goalsAfter"],
#             "tactic": accumulated,
#         }

#         merged_intervals.append(merged_interval)
#         i = j

#     # Transfer trailing `by`
#     for k in range(len(merged_intervals) - 1):
#         if ends_with_by(merged_intervals[k]["tactic"]):
#             txt = merged_intervals[k]["tactic"]
#             cut = txt.rfind(":=") + 2  # keep the ':='
#             by_part = txt[cut:]  # 'by' or ' by', incl. space
#             merged_intervals[k]["tactic"] = txt[:cut]
#             merged_intervals[k + 1]["tactic"] = (
#                 by_part + merged_intervals[k + 1]["tactic"]
#             )

#     # Set goalsBefore of tactic n to be the first non-empty goalsBefore starting at n.
#     for i in range(len(merged_intervals) - 1):
#         new_goals = []
#         for j in range(i, len(merged_intervals)):
#             if merged_intervals[j]["goalsBefore"] != []:
#                 new_goals = merged_intervals[j]["goalsBefore"]
#                 break
#         merged_intervals[i]["goalsBefore"] = new_goals

#     # Set goalsAfter of tactic n to be the goalsBefore of tactic n+1.
#     for i in range(len(merged_intervals) - 1):
#         new_goals = []
#         for j in range(i + 1, len(merged_intervals)):
#             if merged_intervals[j]["goalsBefore"] != []:
#                 new_goals = merged_intervals[j]["goalsBefore"]
#                 break
#         merged_intervals[i]["goalsAfter"] = new_goals

#     return merged_intervals


# def extract_data(infotree: list[dict], source_code: str) -> list[dict]:
#     """
#     Performs the whole extraction process from an infotree and the corresponding Lean4 code.

#     This function:
#       - Extracts nodes and edges from the infotree,
#       - Removes synthetic nodes,
#       - Builds and adjusts intervals to partition the Lean4 code,
#       - Retrieves tactics from the Lean4 code,
#       - Transfers trailing whitespaces and comments between consecutive tactics,
#       - Merge intervals that are not balanced or contain specific tactics.

#     Parameters
#     ----------
#     infotree : list of dict
#         A list of dictionaries each containing 'node' and optionally 'children'.
#     source_code : str
#         The Lean4 code for retrieving text snippets.

#     Returns
#     -------
#     intervals : list of dict
#         A list of dictionaries, each containing:
#           goalsBefore, goalsAfter, tactic.
#         The intervals are partitioned and cover the whole file.
#     """
#     # 1. Extract nodes and edges
#     nodes, _, _ = extract_nodes_and_edges(
#         infotree, include_failed_pp=False, deduplicate=True
#     )

#     # 2. Filter out the synthetic nodes
#     nodes = {
#         k: v
#         for k, v in nodes.items()
#         if not v.get("stx", {}).get("range", {}).get("synthetic", False)
#     }

#     # 3. Build raw intervals from nodes
#     intervals = get_intervals(nodes)

#     # 4. Adjust intervals so they become disjoint and partition the proof
#     intervals = adjust_intervals(intervals)

#     # 5. Load lines from the Lean file
#     source_lines = source_code.split("\n")
#     source_lines = [line + "\n" for line in source_lines[:-1]] + [source_lines[-1]]

#     # 6. Extract the tactic for each final interval
#     intervals = retrieve_tactics(intervals, source_lines)

#     # 7. Transfer trailing whitespaces and comments
#     transfer_trailing_whitespaces_and_comments(intervals)

#     # 8. Merge intervals that are not balanced or contain specific tactics
#     intervals = merge_intervals(intervals)

#     return intervals



import logging
import sys
from typing import Any

from .async_client import AsyncKiminaClient
from .models import (
    BackwardResponse,
    CheckRequest,
    CheckResponse,
    Code,
    Command,
    CommandResponse,
    Diagnostics,
    Error,
    ExtendedCommandResponse,
    ExtendedError,
    Infotree,
    Message,
    ReplRequest,
    ReplResponse,
    Snippet,
    SnippetAnalysis,
    SnippetStatus,
    VerifyRequestBody,
    VerifyResponse,
)
from .sync_client import KiminaClient

__all__ = [
    "AsyncKiminaClient",
    "BackwardResponse",
    "ReplRequest",
    "ReplResponse",
    "CheckRequest",
    "CheckResponse",
    "Code",
    "Command",
    "CommandResponse",
    "Diagnostics",
    "Error",
    "ExtendedCommandResponse",
    "ExtendedError",
    "Infotree",
    "KiminaClient",
    "Message",
    "Snippet",
    "SnippetAnalysis",
    "SnippetStatus",
    "VerifyRequestBody",
    "VerifyResponse",
]

from colorama import Fore, Style, init

init(autoreset=True)


class ColorFormatter(logging.Formatter):
    COLORS = {
        logging.DEBUG: Fore.CYAN,
        logging.INFO: Fore.WHITE,
        logging.WARNING: Fore.YELLOW,
        logging.ERROR: Fore.RED,
        logging.CRITICAL: Fore.MAGENTA + Style.BRIGHT,
    }

    def format(self, record: Any) -> str:
        log_color = self.COLORS.get(record.levelno, "")
        message = super().format(record)
        return f"{log_color}{message}{Style.RESET_ALL}"


logger = logging.getLogger("kimina-client")

if not logger.handlers:
    handler = logging.StreamHandler(sys.stdout)
    formatter = ColorFormatter("%(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)



import logging
from difflib import get_close_matches

from colorama import Style

logger = logging.getLogger("kimina-client")


def find_id_column(columns: list[str]) -> str | tuple[str, str]:
    """
    Finds column that's closest to "id".
    Option to concatenate two column to create id.
    """
    if "id" in columns:
        return "id"

    preferred_names = ["uuid", "proof_id", "problem_id"]

    match = get_close_matches("id", columns, n=1, cutoff=0.6)
    if not match:
        for name in preferred_names:
            match = get_close_matches(name, columns, n=1, cutoff=0.6)
            if match:
                break
    selected_column = match[0] if match else None
    logger.info("Available columns:")
    for i, col in enumerate(columns):
        logger.info(f"{i}: {col}")

    user_input = input("Select column index/name or type 'concat': ").strip()
    if user_input.lower() == "concat":
        idxs = input("Enter two column indices to concatenate (e.g. 0 1): ")
        idx1, idx2 = map(int, idxs.split())
        return (columns[idx1], columns[idx2])
    if not user_input and selected_column:
        return selected_column
    if user_input.isdigit() and 0 <= int(user_input) < len(columns):
        return columns[int(user_input)]
    if user_input not in columns:
        raise ValueError(f"Invalid column: {user_input}")

    return user_input


def find_code_column(columns: list[str]) -> str:
    """
    Finds column with Lean 4 code snippets.
    """
    if "code" in columns:
        return "code"

    preferred_names = ["code", "proof", "full_proof"]

    match = get_close_matches("code", columns, n=1, cutoff=0.6)
    if not match:
        for name in preferred_names:
            match = get_close_matches(name, columns, n=1, cutoff=0.6)
            if match:
                break
    selected_column = match[0] if match else None
    logger.info("Available columns:")
    for i, col in enumerate(columns):
        logger.info(f"{i}: {col}")

    user_input = input("Select column index/name: ").strip()
    if not user_input and selected_column:
        return selected_column
    if user_input.isdigit() and 0 <= int(user_input) < len(columns):
        return columns[int(user_input)]
    if user_input not in columns:
        raise ValueError(f"Invalid column: {user_input}")

    return user_input
    return user_input


def build_log(dataset_name: str, n: int, batch_size: int) -> str:
    """
    String builder to announce benchmark run.
    """
    final_log = (
        f"Running benchmark on {b(dataset_name)}: # Snippets = {b(str(n))} | Batches = "
    )

    n_full_batches = n // batch_size

    if n_full_batches == 0:
        final_log += f"[{b(str(n))}]"
    else:
        if n_full_batches == 1:
            final_log += f"[{b(str(batch_size))}]"
        else:
            final_log += f"{b(str(n_full_batches))} x [{b(str(batch_size))}]"
        if n % batch_size > 0:
            final_log += f" + [{b(str(n % batch_size))}]"
    return final_log


def b(s: str) -> str:
    return str(Style.BRIGHT + s + Style.RESET_ALL)



import json
import logging
import shutil
import textwrap
from enum import Enum
from itertools import chain
from textwrap import wrap
from typing import Any, Literal, NotRequired, Type, TypedDict, TypeVar
from uuid import uuid4

import pygments
from colorama import Fore
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pygments.formatters import Terminal256Formatter
from pygments.lexers import JsonLexer  # type: ignore
from tabulate import tabulate  # type: ignore

logger = logging.getLogger("kimina-client")


class SnippetStatus(str, Enum):
    valid = "valid"
    sorry = "sorry"
    lean_error = "lean_error"  # Error in snippet, clearly identified by message of severity "error"
    repl_error = "repl_error"  # Error while running snippet, at REPL level
    timeout_error = "timeout_error"  # Error caught at server level, which contains "timed out" in the error message
    server_error = (
        "server_error"  # Error caught at server level, which is not a timeout error
    )

    # TODO: fix timing issue related to header non import?


class SnippetAnalysis(BaseModel):
    status: SnippetStatus
    time: float | None = None


class Infotree(str, Enum):
    full = "full"
    tactics = "tactics"
    original = "original"
    substantive = "substantive"


# TODO: Separate schemas in schemas dir with separate files.
class Code(BaseModel):
    custom_id: str | int
    proof: str | None = Field(None)
    code: str | None = Field(
        None
    )  # To be backward compatibility with autoformalizer client

    def get_proof_content(self) -> str:
        content = self.proof if self.proof is not None else self.code
        if content is None:
            raise ValueError(f"Snippet {self.custom_id!r} has no proof/code content")
        return content


class VerifyRequestBody(BaseModel):
    codes: list[Code]
    timeout: int = 300
    infotree_type: Infotree | None = None
    disable_cache: bool = False


class Snippet(BaseModel):
    id: str = Field(..., description="Identifier to trace the snippet")
    code: str = Field(..., description="Lean 4 snippet or proof attempt")

    @classmethod
    def from_code(cls, code: str) -> "Snippet":
        return cls(id=uuid4().hex, code=code)

    @classmethod
    def from_snip(cls, snip: "str | Snippet") -> "Snippet":
        if isinstance(snip, str):
            return cls.from_code(snip)
        return snip


# The classes below map to the REPL/JSON.lean in the Lean REPL repository:
# see https://github.com/leanprover-community/repl


class Command(TypedDict):
    cmd: str
    env: NotRequired[int | None]
    infotree: NotRequired[Infotree]
    gc: NotRequired[bool]


class Pos(TypedDict):
    line: int
    column: int


class Sorry(TypedDict):
    pos: Pos
    endPos: Pos
    goal: str
    proofState: NotRequired[int | None]


class Error(TypedDict):
    message: str


class ExtendedError(Error):
    time: float | None


class Message(TypedDict):
    severity: Literal["trace", "info", "warning", "error"]
    pos: Pos
    endPos: NotRequired[Pos | None]
    data: str


class ProofStep(TypedDict):
    proofState: int
    tactic: str


class Tactic(TypedDict):
    pos: int
    endPos: int
    goals: str
    tactic: str
    proofState: NotRequired[int | None]
    usedConstants: NotRequired[list[str]]


class Diagnostics(TypedDict, total=False):
    repl_uuid: str
    cpu_max: float
    memory_max: float


# TODO: use basemodel pydantic instead
class CommandResponse(TypedDict):
    env: NotRequired[
        int | None
    ]  # Have to make it not required now due to "gc" option already used on previous server
    messages: NotRequired[list[Message] | None]
    sorries: NotRequired[list[Sorry] | None]
    tactics: NotRequired[list[Tactic] | None]
    infotree: NotRequired[Any]


# TODO: write a test validating partition of is_error + is_sorry + is_valid.
# Write the truth table.
def is_error(response: CommandResponse) -> bool:
    messages = response.get("messages")
    if messages is None:
        return False

    return any(m["severity"] == "error" for m in messages)


def has_sorry(response: CommandResponse) -> bool:
    sorries = response.get("sorries")
    if sorries is None:
        return False
    return len(sorries) > 0


def is_valid(response: CommandResponse) -> bool:
    """
    Valid means no error and no sorry.
    """
    return not is_error(response) and not has_sorry(response)


def is_sorry(response: CommandResponse) -> bool:
    """
    Information about sorry shows up twice:
    - as an element in the `sorries` array
    - as a "warning" message with data `declaration uses 'sorry'`

    We rely on the `sorries` array.
    Having a non-empty `sorries` array is not enough to be a sorry.
    You also need to have no errors.
    """
    return not is_error(response) and has_sorry(response)


class ExtendedCommandResponse(CommandResponse):
    time: float | None


# def make_extended_response(
#     response: CommandResponse, time: float | None
# ) -> ExtendedCommandResponse:
#     return ExtendedCommandResponse(**response, time=time)


# def make_extended_error(error: Error, time: float | None) -> ExtendedError:
#     return ExtendedError(**error, time=time)


T = TypeVar("T", bound="ReplRequest")
U = TypeVar("U", bound="ReplResponse")
TS = TypeVar("TS", bound="CheckRequest")
US = TypeVar("US", bound="CheckResponse")


class BaseRequest(BaseModel):
    timeout: int = Field(
        30, description="Maximum time in seconds before aborting the check", ge=0
    )
    debug: bool = Field(
        False, description="Include CPU/RAM usage and REPL instance ID in the response"
    )
    reuse: bool = Field(
        True, description="Whether to attempt using a REPL if available"
    )
    infotree: Infotree | None = Field(
        None,
        description="Level of detail for the info tree.",
    )


class ReplRequest(BaseRequest):
    snippet: Snippet = Field(description="Single snippet to validate")

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "snippet": {
                    "id": "mathlib-import-def",
                    "code": "import Mathlib\ndef f := 1",
                },
                "timeout": 20,
                "debug": False,
                "reuse": True,
                "infotree": "original",
            },
        }
    )


class ReplResponse(BaseModel):
    id: str = Field(..., description="Identifier to trace the snippet")
    time: float = 0.0
    error: str | None = None
    response: CommandResponse | Error | None = None
    diagnostics: Diagnostics | None = None

    def __repr__(self) -> str:
        data = self.model_dump(exclude_none=True)
        json_str = json.dumps(data, indent=2)

        colored: str = pygments.highlight(  # type: ignore
            json_str,
            JsonLexer(),  # type: ignore
            Terminal256Formatter(style="monokai", full=False),  # type: ignore
        ).rstrip()  # type: ignore
        indented = textwrap.indent(colored, 2 * " ")  # type: ignore
        return f"{self.__class__.__name__}(\n{indented}\n)"

    @model_validator(mode="before")
    @classmethod
    def require_error_or_response(
        cls: Type[U], values: dict[str, Any]
    ) -> dict[str, Any]:
        if not values.get("error") and (values.get("response") is None):
            raise ValueError("Either `error` or `response` must be set")
        if values.get("error") is not None and values.get("response") is not None:
            raise ValueError("Only one of `error` or `response` can be set")
        return values

    def analyze(self) -> SnippetAnalysis:
        if self.error is not None:
            if "timed out" in self.error:
                return SnippetAnalysis(status=SnippetStatus.timeout_error)
            return SnippetAnalysis(status=SnippetStatus.server_error)

        if self.response is None:
            raise ValueError(
                f"`ReplResponse` for ID {self.id!r} has no response or error, which should not happen. Please report."
            )

        if "message" in self.response:
            return SnippetAnalysis(status=SnippetStatus.repl_error, time=self.time)

        if is_error(self.response):
            return SnippetAnalysis(status=SnippetStatus.lean_error, time=self.time)
        if is_valid(self.response):
            return SnippetAnalysis(status=SnippetStatus.valid, time=self.time)
        if is_sorry(self.response):
            return SnippetAnalysis(status=SnippetStatus.sorry, time=self.time)

        raise ValueError(
            f"`CommandResponse` for ID {self.id!r} is neither valid, nor sorry, nor error, which should not happen. Please report."
        )


class CheckRequest(BaseRequest):
    snippets: list[Snippet] = Field(
        description="List of snippets to validate (batch or single element)"
    )

    @model_validator(mode="after")
    def check_snippets(self) -> "CheckRequest":
        if not self.snippets:
            raise ValueError("`snippets` must be non empty")

        ids = set({s.id for s in self.snippets})
        if len(ids) != len(self.snippets):
            raise ValueError("`snippets` must have unique ids")
        return self

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "snippets": [
                    {
                        "id": "mathlib-import-def",
                        "code": "import Mathlib\ndef f := 1",
                    },
                ],
                "timeout": 20,
                "debug": False,
                "reuse": True,
                "infotree": "original",
            },
        }
    )


def add_color(text: str, color: str) -> str:
    return str(color + text + Fore.RESET)


def add_percent(count: int, total: int) -> str:
    if count == 0:
        return f"{count}"
    pct = 100 * (count / total)
    if pct >= 10:
        pct_str = f"{int(pct)}"
    elif pct >= 1:
        pct_str = f"{pct:.1f}".rstrip("0").rstrip(".")
    else:
        pct_str = f"{pct:.2f}".rstrip("0").rstrip(".")
    return f"{count} ({pct_str} %)"


def print_summary(
    n: int,
    valid_count: int,
    sorry_count: int,
    lean_error_count: int,
    timeout_error_count: int,
    repl_error_count: int,
    server_error_count: int,
    total_cpu_time: float,
    avg_cpu_time: float,
    elapsed: float,
) -> None:
    n_str = add_color(str(n), Fore.WHITE)
    valid_count_str = add_color(add_percent(valid_count, n), Fore.GREEN)
    sorry_count_str = add_color(add_percent(sorry_count, n), Fore.YELLOW)
    lean_error_count_str = add_color(add_percent(lean_error_count, n), Fore.RED)
    timeout_error_count_str = add_color(
        add_percent(timeout_error_count, n), Fore.MAGENTA
    )
    repl_error_count_str = add_color(add_percent(repl_error_count, n), Fore.RED)
    server_error_count_str = add_color(add_percent(server_error_count, n), Fore.RED)
    total_cpu_time_str = add_color(f"{total_cpu_time:.2f} s", Fore.CYAN)
    avg_cpu_time_str = add_color(f"{avg_cpu_time:.2f} s/snippet", Fore.CYAN)
    elapsed_str = add_color(f"{elapsed:.2f} s", Fore.WHITE)
    table = [
        [
            n_str,
            valid_count_str,
            sorry_count_str,
            lean_error_count_str,
            timeout_error_count_str,
            repl_error_count_str,
            server_error_count_str,
            total_cpu_time_str,
            avg_cpu_time_str,
            elapsed_str,
        ]
    ]
    headers = [
        add_color("#", Fore.WHITE),
        add_color("Valid ✅", Fore.GREEN),
        add_color("Sorry ⚠️", Fore.YELLOW),
        add_color("Lean Error ❌", Fore.RED),
        add_color("Timeout ⏰", Fore.MAGENTA),
        add_color("REPL Error", Fore.RED),
        add_color("Server Error", Fore.RED),
        add_color("Total CPU Time", Fore.CYAN),
        add_color("Avg CPU Time", Fore.CYAN),
        add_color("Elapsed", Fore.WHITE),
    ]

    logger.info(
        "\n"
        + tabulate(table, headers=headers, tablefmt="fancy_grid", stralign="center")  # type: ignore
    )


def log_table_multiline(table_str: str) -> None:
    width = shutil.get_terminal_size((80, 20)).columns
    for line in table_str.splitlines():
        if len(line) > width:
            for chunk in wrap(line, width):
                logger.info(chunk)
        else:
            logger.info(line)


class CheckResponse(BaseModel):
    results: list[ReplResponse]

    @classmethod
    def merge(cls, responses: list["CheckResponse"]) -> "CheckResponse":
        return cls(results=list(chain.from_iterable(r.results for r in responses)))

    def analyze(self, elapsed: float) -> None:
        analyses = [r.analyze() for r in self.results]

        valid_count = sum(1 for a in analyses if a.status == SnippetStatus.valid)
        sorry_count = sum(1 for a in analyses if a.status == SnippetStatus.sorry)
        lean_error_count = sum(
            1 for a in analyses if a.status == SnippetStatus.lean_error
        )
        repl_error_count = sum(
            1 for a in analyses if a.status == SnippetStatus.repl_error
        )
        timeout_error_count = sum(
            1 for a in analyses if a.status == SnippetStatus.timeout_error
        )
        server_error_count = sum(
            1 for a in analyses if a.status == SnippetStatus.server_error
        )

        cpu_times = [a.time for a in analyses if a.time is not None]
        total_cpu_time = sum(cpu_times)
        avg_cpu_time = total_cpu_time / len(cpu_times) if cpu_times else 0.0

        print_summary(
            n=len(self.results),
            valid_count=valid_count,
            sorry_count=sorry_count,
            lean_error_count=lean_error_count,
            timeout_error_count=timeout_error_count,
            repl_error_count=repl_error_count,
            server_error_count=server_error_count,
            total_cpu_time=total_cpu_time,
            avg_cpu_time=avg_cpu_time,
            elapsed=elapsed,
        )


class BackwardResponse(TypedDict):
    custom_id: str
    error: NotRequired[
        str | None
    ]  # TODO: check if error is required here, probably not
    response: NotRequired[ExtendedCommandResponse | ExtendedError | None]


def backward_response_from_repl(repl_response: ReplResponse) -> BackwardResponse:
    data = BackwardResponse(custom_id=repl_response.id)
    if repl_response.error is not None:
        data["error"] = repl_response.error
    if repl_response.response is not None:
        data["response"] = extend(repl_response.response, time=repl_response.time)
    return data


def extend(
    response: CommandResponse | Error | None, time: float | None = None
) -> ExtendedCommandResponse | ExtendedError | None:
    if response is None:
        return None
    elif "message" in response:
        return ExtendedError(**response, time=time)  # type: ignore
    else:
        return ExtendedCommandResponse(**response, time=time)


class VerifyResponse(BaseModel):
    results: list[BackwardResponse]

    def __repr__(self) -> str:
        data = self.model_dump(exclude_none=True)
        json_str = json.dumps(data, indent=2)

        colored = pygments.highlight(  # type: ignore
            json_str,
            JsonLexer(),  # type: ignore
            Terminal256Formatter(style="monokai", full=False),  # type: ignore
        ).rstrip()

        indented = textwrap.indent(colored, "  ")  # type: ignore
        return f"{self.__class__.__name__}(\n{indented}\n)"



import re
import statistics
from dataclasses import dataclass
from typing import Any, TypedDict

from .models import (
    BackwardResponse,
    CommandResponse,
    Error,
    ExtendedCommandResponse,
    ExtendedError,
    Message,
    Pos,
)

_ERR_RE = re.compile(r"^([^:]+):\n(.*)", re.DOTALL)


class FinalMessage(TypedDict):
    severity: str
    message: str
    pos: Pos
    endPos: Pos | None


def parse_messages(messages: list[Message]) -> list[FinalMessage]:
    parsed_messages: list[FinalMessage] = []
    for msg in messages:
        parsed_messages.append(
            FinalMessage(
                severity=msg.get("severity", "info"),
                message=msg.get("data", ""),
                pos=msg.get("pos", Pos(line=0, column=0)),
                endPos=msg.get("endPos", Pos(line=0, column=0)),
            )
        )
    return parsed_messages


def parse_error_message(message: str) -> list[FinalMessage]:
    # TODO : compare with partition c call and add real-life tests
    m = _ERR_RE.match(message)
    if m:
        severity, text = m.groups()
    else:
        severity, text = "error", message
    return [
        FinalMessage(
            severity=severity,
            message=text,
            pos=Pos(line=0, column=0),
            endPos=Pos(line=0, column=0),
        )
    ]


def parse_lean_response(response: CommandResponse | Error) -> dict[int, FinalMessage]:
    messages: list[FinalMessage] = []
    if "messages" in response:
        messages = parse_messages(response.get("messages", []) or [])  # type: ignore
    elif "message" in response:
        messages = parse_error_message(response.get("message", ""))  # type: ignore

    # TODO: @marco is it ok to filter out unsolved goals?
    # messages = list(filter(lambda x: "unsolved goals" not in x["message"], messages))
    # messages = sorted(messages, key=lambda x: (x["pos"]["line"], x["pos"]["column"]))

    # here if multiple errors on same line it will take last, why not first?
    # line_num_to_message = {(message["pos"]["line"]): message for message in messages}
    # here if multiple errors on same line it will take first

    # TODO: add tests on this
    # dict comprehension keeps last value for each key, so reverse gives first
    return {m["pos"]["line"]: m for m in reversed(messages)}


def get_messages_for_lines(
    messages: dict[int, FinalMessage], start_line: int, end_line: int
) -> tuple[list[FinalMessage], bool, bool]:
    selected_messages: list[FinalMessage] = []
    has_error = False
    is_unsolved_goals = False
    for idx in range(start_line, end_line + 1):
        if idx in messages:
            selected_messages.append(
                messages[idx]
            )  # TODO: check how we ensure there is indeed a message at line idx?
            if messages[idx]["severity"] == "error":
                has_error = True
            if "unsolved goals" in messages[idx]["message"]:
                is_unsolved_goals = True
    return selected_messages, has_error, is_unsolved_goals


# TODO: check who uses this apart from tests here.
def has_error_response(
    feedback: ExtendedCommandResponse | ExtendedError | None,
    accept_sorry: bool = True,
    return_error_messages: bool = False,
) -> bool:
    """
    Checks if the Lean feedback contains an error.

    Args:
        feedback: The Lean feedback as a dictionary.
        accept_sorry: Whether to accept "sorry" statements as "not an error".
            By default, "sorry" statements are not considered errors.
        return_error_messages: Whether to return the feedback error messages.
    """

    # TODO: check. This is a bit weird here, there is never error or stderr in the feedback...
    # if "error" in feedback:
    #     r = (True, [feedback["error"]]) if return_error_messages else True
    #     return r

    # if "stderr" in feedback:
    #     r = (True, [feedback["stderr"]]) if return_error_messages else True
    #     return r

    has_error = False
    error_data_values = []
    sorry_data_values = []
    if "messages" in feedback:  # type: ignore
        error_data_values = [
            message["data"]
            for message in feedback.get("messages", [])  # type: ignore
            if message.get("severity") == "error"
        ]
        has_error = bool(error_data_values)

        if not accept_sorry:
            warning_data_values = [
                message["data"]
                for message in feedback.get("messages", [])  # type: ignore
                if message.get("severity") == "warning"
            ]
            sorry_data_values = [
                warning_data
                for warning_data in warning_data_values
                if "declaration uses 'sorry'" in warning_data
            ]
            has_error = has_error or bool(sorry_data_values)

    if return_error_messages:
        return has_error, error_data_values + sorry_data_values  # type: ignore
    else:
        return has_error


class ParsedClientResponse(TypedDict):
    has_error: bool
    is_valid_no_sorry: bool
    is_valid_with_sorry: bool
    time: float | None


# TODO: Remove all this code
def parse_client_response(response: BackwardResponse) -> ParsedClientResponse:
    """
    Parses the response from the Lean4Client.
    Reponse should be the output of client.Lean4Client.(async_)verify which is BackwardResponse.
    TODO: Make sure that we implement the proper function for CheckResponse

    Args:
        - response (BackwardResponse): The response from the Lean4Client.

    Returns:
        - dict: A dictionary containing the following keys:
            - has_error: Whether the response contains an error.
            - is_valid_no_sorry: Whether the response is valid without "sorry" statements.
                this is used for proof verification.
            - is_valid_with_sorry: Whether the response is valid with "sorry.
                this is used for statement verification.
    """
    error_message = response.get("error", None)
    json_response: Any = response.get(
        "response", {}
    )  # time used to be included in the lean response :/

    error = bool(error_message) or has_error_response(json_response)  # type: ignore
    is_valid_no_sorry = (not bool(error_message)) and (
        not has_error_response(json_response, accept_sorry=False)  # type: ignore
    )
    is_valid_with_sorry = (not bool(error_message)) and (
        not has_error_response(json_response, accept_sorry=True)  # type: ignore
    )

    return ParsedClientResponse(
        has_error=error,
        is_valid_no_sorry=is_valid_no_sorry,
        is_valid_with_sorry=is_valid_with_sorry,
        time=json_response.get("time", None) if json_response else None,
    )


@dataclass
class SampleAnalysis:
    is_valid_no_sorry: bool
    is_valid_with_sorry: bool
    has_error: bool
    has_connection_error: bool
    time: float | None


# Only used here locally in this repo.
def analyze_sample(lean_feedback: BackwardResponse) -> SampleAnalysis:
    error = lean_feedback.get("error", None)
    output = parse_client_response(lean_feedback)

    return SampleAnalysis(
        is_valid_no_sorry=output["is_valid_no_sorry"],
        is_valid_with_sorry=output["is_valid_with_sorry"],
        has_error=output["has_error"],
        has_connection_error=bool(error),
        time=None if (error is not None and "timed out" in error) else output["time"],
    )


# Move to benchmark maybe
def analyze(results: list[BackwardResponse]) -> None:
    analyses = [analyze_sample(sample_result) for sample_result in results]
    total = len(analyses)

    valid_count = sum(a.is_valid_no_sorry for a in analyses)
    conn_errors = sum(a.has_connection_error for a in analyses)
    times = [a.time for a in analyses if a.time is not None]
    timeouts = total - len(times)

    print(f"Valid proofs: {100 * (valid_count / total):.2f} %")
    print(f"Connection errors rate: {100 * (conn_errors / total):.2f} %")

    # TODO: put the results of the analysis in an object which has a proper __repr__
    print(
        f"Total verification time: {sum(times):.2f} seconds "
        f"(excluding {timeouts} timeout{'' if timeouts == 1 else 's'})"
    )

    if times:
        print(
            f"Average verification time: {statistics.mean(times):.2f} seconds per successful verification"
        )
